/*
    Copyright (c) 2002 Jens Tinz
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
    1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
    3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/**
 * @fileoverview This module defines the menu's actions and layout.
 *
 * To use this construct an instance of {@link ActionManager} and then use
 * manager.layouts. You should not store your own reference to
 * manager.layouts since it is replaced on preference changes. Do not
 * forget to call shutdown to unregister the preference observer.
 *
 * To add standard actions add them to _simpleActions. To add
 * nonstandard actions edit the {@link ActionManager} constructor. To add or
 * modify layouts edit {@link ActionManager#_updateLayouts}.
 *
 * See integration.js for the normal value of the context argument.
 * When customizing the context is null. Tweak functions must handle this.
 */

// XXX _simpleActions above should be a @link but jsdoc-1.10.2 does
// not pick up var documentation correctly.

const menu = Components.utils.import(
    'resource://radialcontext-mz/menu.js', null);
const utils = Components.utils.import(
    'resource://radialcontext-mz/utils.js', null);

const bundle = Components.classes['@mozilla.org/intl/stringbundle;1']
    .getService(Components.interfaces.nsIStringBundleService)
    .createBundle('chrome://radialcontext-mz/locale/labels.properties');
/**
 * Gettext-ish function.
 *
 * When called with one argument this is just GetStringFromName.
 * When called with more arguments it is formatStringFromName (with varargs).
 *
 * @param {String} label name to pass to the stringbundle.
 */
function _(label) {
    if (arguments.length === 1)
        return bundle.GetStringFromName(label);
    return bundle.formatStringFromName(
        label,
        Array.prototype.slice.call(arguments, 1), arguments.length - 1);
}

/**
 * Map from action name to action function.
 * The name is used as-is as label name and with the first "_"
 * replaced with a "/" as icon name.
 */
const _simpleActions = {
    // Keep these grouped by submenu.

    navigation_up: function(context) {
        // location.href works on error pages
        // (document.URL is about:neterror?...)
        let url = context.window.content.location.href;
        url = url.slice(0, url.slice(0, -1).lastIndexOf('/') + 1);
        context.window.loadURI(url);
    },
    navigation_firstpage: function(context) {
        // code adapted from EasyGestures
        context.window.getWebNavigation().gotoIndex(0);
    },
    navigation_lastpage: function(context) {
        // code adapted from EasyGestures
        let nav = context.window.getWebNavigation();
        nav.gotoIndex(nav.sessionHistory.count - 1);
    },

    tab_duplicate: function(context) {
        context.window.openNewTabWith(context.window.content.location.href,
                                      null, null, null, false);
    },
    tab_next: function(context) {
        context.window.getBrowser().tabContainer.selectedIndex++;
    },
    tab_prev: function(context) {
        context.window.getBrowser().tabContainer.selectedIndex--;
    },
    tab_close_others: function(context) {
        let browser = context.window.getBrowser();
        // Same function browser/base/content/tabbrowser.xml uses.
        browser.removeAllTabsBut(browser.selectedTab);
    },

    window_duplicate: function(context) {
        context.window.open(context.window.content.location);
    },
    window_close_others: function(context) {
        // adapted from easyGestures
        let enumerator = Components.classes[
            '@mozilla.org/appshell/window-mediator;1']
            .getService(Components.interfaces.nsIWindowMediator)
            .getEnumerator('navigator:browser');
        while (enumerator.hasMoreElements()) {
            let awin = enumerator.getNext();
            if (awin !== context.window)
                awin.close();
        }
    },
    window_minimize: function(context) {
        // The mac-specific (I think) minimizeWindow action uses this too:
        context.window.minimize();
    },

    page_validate: function(context) {
        context.window.loadURI('http://validator.w3.org/check?uri=' +
                               context.window.content.document.URL);
    },

    // I do not see an equivalent to these in the context menu.
    // Not sure if we need anything fancier than what is already here.
    selection_open: function(context) {
        context.window.loadURI(context.selection);
    },
    selection_open_ntab: function(context) {
        context.window.openNewTabWith(context.selection,
                                      null, null, null, true);
    },
    // XXX I dislike having hardcoded urls in here like this.
    // I am really tempted to drop this and tell people to use a search
    // plugin instead (mycroft has babelfish plugins) but I think
    // that would be the first removed feature from the original
    // RadialContext and I do not want to break backwards compat over
    // this.
    selection_translate: function(context) {
        context.window.loadURI('http://babelfish.altavista.com/');
        // XXX setTimeout abuse, should probably use an onload event here.
        context.window.setTimeout(function(content, selection) {
                content.document.getElementsByTagName('textarea')[0].value =
                    selection;
            }, 3000, context.window.content, context.selection);
    },
    selection_translate_ntab: function(context) {
        context.window.openNewTabWith('http://babelfish.altavista.com/',
                                      null, null, null, false);
        // XXX setTimeout abuse, should probably use an onload event here.
        context.window.setTimeout(function(content, selection) {
                content.document.getElementsByTagName('textarea')[0].value =
                    selection;
            }, 3000, context.window.content, context.selection);
    }
};

/**
 * Like _simpleActions but for actions built on top of the standard
 * Firefox context menu.
 */
// Actions last checked with Firefox 3.0.
const _simpleContextMenuActions = {
    frame_view: function(ctxmenu) {
        ctxmenu.showOnlyThisFrame();
    },
    frame_reload: function(ctxmenu) {
        ctxmenu.reloadFrame();
    },
    frame_save: function(ctxmenu) {
        ctxmenu.saveFrame();
    },
    frame_info: function(ctxmenu) {
        ctxmenu.viewFrameInfo();
    },
    frame_source: function(ctxmenu) {
        ctxmenu.viewFrameSource();
    },

    link_in_newtab: function(ctxmenu) {
        ctxmenu.openLinkInTab();
    },
    link_in_newwin: function(ctxmenu) {
        ctxmenu.openLink();
    },
    link_save: function(ctxmenu) {
        ctxmenu.saveLink();
    },
    link_copy_email: function(ctxmenu) {
        ctxmenu.copyEmail();
    },

    image_view: function(ctxmenu) {
        // pass null for the event, which is only used for
        // browser/base/content/utilityOverlay.js:whereToOpenLink,
        // which deals with null.
        ctxmenu.viewImage(null);
    },
    image_save: function(ctxmenu) {
        ctxmenu.saveImage();
    },

    selection_source: function(ctxmenu) {
        ctxmenu.viewPartialSource('selection');
    }
};

/**
 * Map from action name to command id.
 */
const _commandActions = {
    navigation_stop: 'Browser:Stop',
    navigation_reload: 'Browser:Reload',
    navigation_reload_skip: 'Browser:ReloadSkipCache',
    navigation_enter_address: 'Browser:OpenLocation',

    tab_new: 'cmd_newNavigatorTab',
    tab_close: 'cmd_close',
    tab_undo_close: 'History:UndoCloseTab',

    window_new: 'cmd_newNavigator',
    window_close: 'cmd_closeWindow',
    window_fullscreen: 'View:FullScreen',
    window_togglebar: 'viewBookmarksSidebar',

    page_bookmark: 'Browser:AddBookmarkAs',
    page_save: 'Browser:SavePage',
    page_source: 'View:PageSource',
    page_info: 'View:PageInfo'
};

/**
 * Map from action name to command name for controller-implemented commands.
 */
const _controllerCommandActions = {
    link_copy_url: 'cmd_copyLink',

    image_copy_url: 'cmd_copyImageLocation',

    // Same action edit_copy uses, but different icon.
    selection_copy: 'cmd_copy',

    edit_cut: 'cmd_cut',
    edit_copy: 'cmd_copy',
    edit_paste: 'cmd_paste',
    edit_undo: 'cmd_undo'
};


// These save typing/repetition and provide a name attribute options.js uses.
// Be careful: options.js relies on that name attribute, so everything here
// *must* use Localized* instead of menu.* or set the name by hand!

function LocalizedAction(name) {
    // Yes, this only replaces the first one found
    menu.Action.call(this, _(name), name.replace('_', '/'));
    this.name = name;
}
LocalizedAction.prototype = menu.Action.prototype;

function LocalizedSubmenu(name, layout) {
    menu.Submenu.call(this, _(name), name.replace('_', '/'), layout);
    this.name = name;
}
LocalizedSubmenu.prototype = menu.Submenu.prototype;

function LocalizedDummy(name) {
    menu.Dummy.call(this, _(name));
    this.name = name;
}
LocalizedDummy.prototype = menu.Dummy.prototype;


// I apparently do not understand commands.
//
// There are apparently two kinds of command: the kind you should use
// by invoking goDoCommand and the kind you should use by invoking
// getElementById(...).doCommand(). And occasionally you can do both,
// because there is a <command id="cmd_paste"
// oncommand="goDoCommand('cmd_paste')"/>, so the getElementById one
// invokes to the goDoCommand one. Only in that case the
// getElementById one may not get its enabled state updated along with
// the goDoCommand one, so you should probably still use the
// goDoCommand one to be safe (I do not know why. Perhaps there is no
// general way to get a notification when the goDoCommand command's
// enabled state changes? Or perhaps they were just too lazy to add
// more than that single line of wrapper?).

/**
 * Set and restore context.window.document.popupNode around a function call.
 *
 * @param self Passed as "this" to the "func" argument.
 * @param {function} func The function to call.
 * @param context Normal menu context. "window" and "target" attrs are used.
 */
function _withPopupNode(self, context, func) {
    // I am not entirely sure this is sane, but without it the
    // nsContextMenu needs crazy hacking (it accesses this) and a
    // bunch of commands simply do not work (cmd_copyLink for example
    // raises on the isCommandEnabled check (!) from C++ code if this
    // unset).
    let doc = context.window.document;
    let originalPopupNode = doc.popupNode;
    doc.popupNode = context.target || context.frameDoc.body;
    try {
        return func.call(self);
    } finally {
        // I am pretty sure this reset is not actually needed, because
        // if we leave popupNode alone entirely commands accessing it
        // start working if you open the normal context menu on the
        // right node right before opening the popup menu. So I think
        // the normal context menu leaves this set. Still, it seems
        // cleaner to reset this:
        doc.popupNode = originalPopupNode;
    }
}

/**
 * @class Action bound to a command (not implemented by a controller).
 *
 * Automagically gets disabled/enabled based on the command it is bound to.
 * The label from the command is not used. This is necessary to work
 * in the customization dialog.
 *
 * @param {String} name
 * @param {String} command
 *
 */
function CommandAction(name, command) {
    LocalizedAction.call(this, name);
    this.command = command;
}

CommandAction.prototype = {
    __proto__: LocalizedAction.prototype,

    tweak: function(context) {
        if (context === null)
            this.active = true;
        else
            _withPopupNode(
                this, context, function() {
                    this.active = !context.window.document.getElementById(
                        this.command).hasAttribute('disabled');
                });
    },

    doAction: function(context) {
        // I assume we do not need to check for disabled again here.
        // The menu never calls this if we were disabled when tweak()
        // was called, and I assume the underlying command code deals
        // in the rare case the disabled state changed after that.
        _withPopupNode(
            this, context, function() {
                context.window.document.getElementById(this.command)
                    .doCommand();
            });
    }
};

/**
 * @class Action bound to a command implemented by a controller.
 *
 * Automagically gets disabled/enabled based on the command it is bound to.
 * The label from the command is not used. This is necessary to work
 * in the customization dialog.
 *
 * @param {String} name
 * @param {String} command
 */
function ControllerCommandAction(name, command) {
    LocalizedAction.call(this, name);
    this.command = command;
}

ControllerCommandAction.prototype = {
    __proto__: LocalizedAction.prototype,

    tweak: function(context) {
        if (context === null)
            this.active = true;
        else
            _withPopupNode(
                this, context, function() {
                    // Based on goDoCommand from globalOverlay.js
                    // in the Mozilla tree.
                    let controller = context.window.top.document
                        .commandDispatcher
                        .getControllerForCommand(this.command);
                    this.active = (controller &&
                                   controller.isCommandEnabled(this.command));
                });
    },

    doAction: function(context) {
        _withPopupNode(this, context, function() {
                           context.window.goDoCommand(this.command);
                       });
    }
};


/**
 * @class Action using the Mozilla context menu.
 */
function ContextMenuAction(name) {
    LocalizedAction.call(this, name);
}

ContextMenuAction.prototype = {
    __proto__: LocalizedAction.prototype,

    doAction: function(context) {
        _withPopupNode(
            this, context, function() {
                let ctxmenu = new context.window.nsContextMenu(
                    null, context.window.getBrowser());
                this.doMenuAction(ctxmenu);
                // ctxmenu has onDestroy but it is a noop and nothing calls it
                // (unless I screwed up my search). So we don't bother either.
            });
    },

    /**
     * Override this in subclasses to actually do stuff.
     *
     * @param {nsContextMenu} ctxmenu The context menu instance.
     */
    doMenuAction: function(ctxmenu) {}
};


/**
 * @class Helper object used by the scrollMenu.
 *
 * This manages a timer which handles scrolling in a single direction.
 *
 * @implements nsITimerCallback
 *
 * @param {Function} scrollFunc Called with a distance in (integer) pixels.
 */
// Background info on what this does and why:
//
// Mozilla forces the interval on timers to be >= 10 ms. On my windows
// system it seems to be forced to >= 15 ms. Larger intervals give
// more different speeds (speed changes smoother when the mouse
// moves). It also seems that on average (but *only* on average) the
// "sensitivity" is better than 15 ms (at least on the windows system
// I am currently testing on): running a setInterval loop with an
// interval of 20 or 21 seconds 100 times gives an average interval
// within 0.5 ms of the target. The *individual* deltas may or may not
// be any good, since I am measuring with Date.now() and I do not
// trust its resolution. It claims the individual deltas vary from 2
// to 50 ms.
//
// Some (limited) experimenting indicates using setTimeout or running
// a setInterval once then cancelling and reinitializing it with the
// same timeout value has noticable (a few ms) overhead. So try to
// leave the setInterval running if possible.
//
// Scrolling (window.scrollBy) works in integer pixels. We definitely
// need to scroll more than one pixel per interval for higher scroll
// speeds, but to allow really slow scrolling while also keeping fast
// scrolls smooth we *also* vary the scroll interval.
//
// To DTRT when we get a lot (several per scroll interval) of setSpeed
// calls we cannot just cancel the current timeout in setSpeed and set
// a new one because we do not know how far we are from the timeout
// expiring. If we did that we would stop scrolling as long as the
// mouse is moving. We cannot reliably track how far we are from the
// timeout by hand because we cannot query the current time with a
// reasonable resolution (Date.now has a resolution of about 15ms on
// windows on at least some Firefox versions). So we should usually
// wait for the current interval to trigger before setting a new one.
// Because setInterval seems to be more accurate than setTimeout if we
// run it more than once we do want to use setInterval instead of just
// always using setTimeout and unconditionally rescheduling at the end
// of the timeout function.
//
// Most of the time this is a nonissue because the next timeout will
// happen "soon enough" (compared to human reaction times) anyway. But
// if we are currently scrolling *really* slowly and want to speed up
// we need to cancel the timer. Specifically currently
// (minScrollInterval 30 ms, scrollSpeed 0.0002) our slowest scroll
// speeds are 1 pixel in 5 *seconds*, then 1.25 seconds, ~0.6 seconds,
// ~0.3 seconds, etc. If we are in that 5 second timeout cancelling it
// is almost certainly an improvement. So we only reset our interval
// if we are going from really slow to normal or fast scrolling. In
// other cases we wait for the next timer event and reset from there.
//
// Currently the rule used is "if the current timeout is more than
// twice the new timeout we cancel it" (because in that case it is
// fairly likely our new timeout will happen sooner than the one we
// are cancelling). Notice this implies that the previous scroll
// distance is 1 pixel. The next interval is >= minScrollInterval, so
// if currentInterval >= 2 * nextInterval then currentinterval >= 2 *
// minScrollInterval. If we would scroll one more pixel per interval
// the time between intervals would be currentInterval * (1 - 1 / d),
// and if d >= 2 then (1 - 1 / d) >= 1/2, so that would still be >=
// minScrollInterval, so we would be doing that. So we only cancel the
// old timer if it was scrolling at a rate of 1 pixel and either the
// new scroll rate is also one pixel (so even if we accidentally delay
// the next scroll a lot it will not be that noticable) or it has a
// timeout of < 2 * minScrollInterval (so it is fast enough that
// rescheduling is almost certainly "safe", and we will not reschedule
// again until this one runs).
function Scroller(scrollFunc) {
    this.scrollFunc = scrollFunc;
    // New target time in ms between scroll events.
    // Meaningless if this.timer === null.
    this.nextInterval = -1;
    // Distance in integer (!) pixels to scroll per event. Also
    // meaningless if this.timer === null.
    this.distance = 0;
    // The currently running timer (nsITimer).
    // XXX we *should* be able to reuse the timer after cancelling it,
    // but I ran into weird bugs that went away when I dropped the reuse.
    // So for now this is a non-cancelled timer or null.
    this.timer = null;
    // Dynamically adjusted minScrollInterval.
    this.minScrollInterval = this.defaultMinScrollInterval;
    // Counter for a number of loops with the same delay interval.
    this.loops = null;
    // Time (from Date.now()) at which this.loops was set to 0.
    this.startTime = null;
    // startTime and loops are set when the timer delay changes.
}

Scroller.prototype = {
    /**
     * The minimum delay between scroll events (ms per interval). Do not
     * lower this too much: if the browser cannot keep up this causes
     * awkward jumps in scroll speed when we switch from scrolling 1 pixel
     * per interval to scrolling 2 pixels per interval.
     */
    defaultMinScrollInterval: 30,
    /**
     * The maximum delay between scroll events (ms per interval). If the
     * browser cannot keep up with this rate the scroll speed becomes
     * inaccurate. We need this to prevent system load or browser
     * "hickups" from causing insane scroll interval values, especially
     * since we currently cannot lower the min scroll interval.
     */
    maxScrollInterval: 1000,

    /**
     * Update the speed in pixels / ms. Call this with 0 to stop scrolling.
     *
     * Calling with 0 cancels the running timer if there is one. You want
     * to do that before discarding this object.
     *
     * @param {int} pixelsPerMs Desired speed.
     */
    setSpeed: function(pixelsPerMs) {
        if (pixelsPerMs) {
            // We want our pixels per interval to be an integer value. We
            // also want to round it away from zero to make sure the
            // interval we calculate from this is > minScrollInterval.
            // That is *down* (floor instead of ceil) for numbers < 0:
            if (pixelsPerMs > 0)
                this.distance = Math.ceil(
                    this.minScrollInterval * pixelsPerMs);
            else
                this.distance = Math.floor(
                    this.minScrollInterval * pixelsPerMs);
            // Make this an integer because the timer functions take one.

            // Notice that distance has the same sign as pixelsPerMs, so
            // nextInterval is always positive.
            this.nextInterval = Math.ceil(this.distance / pixelsPerMs);
            // Guess that waiting for the current timer to expire would
            // take too long if the current interval is more than twice
            // the new interval.
            if (this.timer !== null &&
                this.timer.delay > 2 * this.nextInterval)
                this._reschedule();
            else if (this.timer === null) {
                this.timer = utils.Timer(this, this.nextInterval,
                                         utils.TIMER_REPEATING_PRECISE);
                this.startTime = Date.now();
                this.loops = 0;
            }
        } else {
            // It is a request to stop scrolling.
            if (this.timer !== null) {
                this.timer.cancel();
                this.timer = null;
            }
        }
    },

    /**
     * Reschedule the timer based on nextInterval.
     */
    _reschedule: function() {
        this.timer.delay = this.nextInterval;
        this.startTime = Date.now();
        this.loops = 0;
    },

    /**
     * Callback function for the timer.
     */
    notify: function(timer) {
        this.scrollFunc(this.distance);
        ++this.loops;
        // We must have had a few scroll events (to average out errors)
        // and we must have spent some time scrolling (because the
        // resolution of Date.now() is not that great)
        let expectedTime = this.loops * this.timer.delay;
        if (this.loops > 10 && expectedTime > 500) {
            let now = Date.now();
            let actualTime = now - this.startTime;
            if (actualTime > 1000) {
                if (actualTime > expectedTime * 1.1) {
                    this.minScrollInterval = actualTime / this.loops;
                    if (this.minScrollInterval > this.maxScrollInterval)
                        this.minScrollInterval = this.maxScrollInterval;
                    // and reschedule the timer if necessary:
                    if (this.minScrollInterval > this.nextInterval)
                        this.setSpeed(this.distance / this.nextInterval);
                }
                this.startTime = now;
                this.loops = 0;
            }
        }
        if (this.timer.delay !== this.nextInterval)
            this._reschedule();
    }
};


/**
 * @class Special submenu triggering scrolling.
 *
 * This "freezes" the menu in place on the screen as an indication of
 * where to move the mouse to stop scrolling. That is unusual enough
 * that the implementation is rather "special".
 */
function ScrollMenu() {
    // Entire menu intentionally hardcoded: no point in making this
    // customizable, and no point in using the scroll_* actions anywhere
    // but here.
    let dummy = new menu.Dummy();
    LocalizedSubmenu.call(this, 'menu_scroll', [
            dummy, new LocalizedAction('scroll_right'),
            dummy, new LocalizedAction('scroll_down'),
            dummy, new LocalizedAction('scroll_left'),
            dummy, new LocalizedAction('scroll_up'),
        ]);
    // These are set from activateMenu and killed when the menu closes.
    this.horizScroller = this.vertScroller = null;
    // Likewise.
    this.targetView = null;
}

ScrollMenu.prototype = {
    __proto__: LocalizedSubmenu.prototype,

    /**
     * With the mouse d pixels over the edge of the menu, scroll speed is
     * d * d * scrollSpeed pixels per millisecond,
     * rounded to whole pixels per interval.
     * So the larger this value, the faster we scroll.
     */
    scrollSpeed: 0.000001,

    activateMenu: function(piemenu) {
        // Hide the labels even in "show always/immediately" mode.
        piemenu.hideLabels();

        // Define our scrollers.
        let targetView = piemenu.context.frameDoc.defaultView;
        this.horizScroller = new Scroller(function (d) {
                targetView.scrollBy(d, 0);
            });
        this.vertScroller = new Scroller(function (d) {
                targetView.scrollBy(0, d);
            });

        // Minor hack: mousemove uses this to disable the scrolling timer
        // when we cannot scroll in that direction.
        this.targetView = targetView;

        // HACK: we completely hijack the PieMenu by replacing its
        // callback methods with ours. That is ugly, but guarantees we do
        // not install a misbehaving event handler if someone is doing
        // something unusual with the menu, and also takes care of
        // stopping the menu from moving around.
        let self = this;
        piemenu.update = function(mx, my) {
            self.mousemove(piemenu, mx, my);
            // Claim we never move so integration never bothers with
            // tooltip timers.
            return false;
        };
        // We still need this override if there is a nonzero tooltip delay
        // and "keep on movement" is enabled.
        piemenu.showLabels = function() {};

        // A bit tricky since we need to call the original method.
        let close = piemenu.close;
        piemenu.close = function(doAction) {
            // Kill the timers and drop the reference.
            self.vertScroller.setSpeed(0);
            self.horizScroller.setSpeed(0);
            self.vertScroller = null;
            self.horizScroller = null;
            self.targetView = null;
            close.call(piemenu, doAction);
        };
    },

    /**
     * Handler for mousemove events.
     *
     * @param {PieMenu} piemenu The menu instance we are overriding.
     * @param {int} mx
     * @param {int} my
     */
    mousemove: function(piemenu, mx, my) {
        // This used to calculate speed the same way the menu calculates
        // movement (with no scrolling while the mouse is in the hole in
        // the menu) but that feels odd in the common case of no
        // horizontal scrolling (moving the mouse horizontally affects
        // vertical scrolling speed). So now we simply calculate
        // horizontal and vertical speed separately.

        // TODO experiment with different functions to get the speed from
        // the distance. The current distance-squared one is pretty random.

        // Possibly unnecessary optimization: we force the scroller off if
        // the corresponding scrollMax value is 0 (if we cannot scroll in
        // that direction). Done because I am not sure how well various
        // Moz versions deal with two timers with a short but nonidentical
        // interval (and it is very common to not have a horizontal
        // scrollbar).
        let d = mx - piemenu.x;
        let d2 = Math.abs(d) - piemenu.theme.actionRadius;
        if (d2 > 0 && this.targetView.scrollMaxX)
            // Cursor outside the menu.
            this.horizScroller.setSpeed((d > 0 ? 1 : -1) *
                                        d2 * d2 * d2 * this.scrollSpeed);
        else
            this.horizScroller.setSpeed(0);

        d = my - piemenu.y;
        d2 = Math.abs(d) - piemenu.theme.actionRadius;
        if (d2 > 0 && this.targetView.scrollMaxY)
            this.vertScroller.setSpeed((d > 0 ? 1 : -1) *
                                       d2 * d2 * d2 * this.scrollSpeed);
        else
            this.vertScroller.setSpeed(0);
    }
};


/**
 * @class Forward button that grays out when at the end of history.
 */
function NaviForward() {
    CommandAction.call(this, 'navigation_forward', 'Browser:Forward');
}

NaviForward.prototype = {
    __proto__: CommandAction.prototype,

    tweak: function(context) {
        CommandAction.prototype.tweak.call(this, context);
        if (this.active)
            this.iconname = 'navigation/forward';
        else
            this.iconname = 'navigation/forward_gray';
    }
};


/**
 * @class Back button that grays out when at the beginning of history.
 */
function NaviBack() {
    CommandAction.call(this, 'navigation_back', 'Browser:Back');
}

NaviBack.prototype = {
    __proto__: CommandAction.prototype,

    tweak: function(context) {
        CommandAction.prototype.tweak.call(this, context);
        if (this.active)
            this.iconname = 'navigation/back';
        else
            this.iconname = 'navigation/back_gray';
    }
};

/**
 * @class Action invoking the selected Mozilla search engine on the selection.
 *
 * @param {bool} inNewTab
 */
function SearchAction(inNewTab) {
    if (inNewTab)
        LocalizedAction.call(this, 'selection_search_ntab');
    else
        LocalizedAction.call(this, 'selection_search');

    this.inNewTab = inNewTab;
}

SearchAction.prototype = {
    __proto__: LocalizedAction.prototype,

    tweak: function(context) {
        if (context === null)
            return;
        // Based on nsContextMenu.js:isTextSelection from 3.0b5.
        let searchService = Components.classes
            ['@mozilla.org/browser/search-service;1']
            .getService(Components.interfaces.nsIBrowserSearchService);
        let searchBar = context.window.BrowserSearch.searchBar;
        let engineName;
        if (searchBar !== null && context.window.isElementVisible(searchBar))
            engineName = searchService.currentEngine.name;
        else
            engineName = searchService.defaultEngine.name;

        if (this.inNewTab)
            this.label = _('selection_search_engine_ntab', engineName);
        else
            this.label = _('selection_search_engine', engineName);
    },

    doAction: function(context) {
        context.window.BrowserSearch.loadSearch(context.selection,
                                                this.inNewTab);
    }
};


/**
 * @class Label displaying the server affected by image blocking.
 */
function ImageServerLabel() {
    LocalizedDummy.call(this, 'image_server');
}

ImageServerLabel.prototype = {
    __proto__: LocalizedDummy.prototype,

    tweak: function(context) {
        // Use label from construction when customizing:
        if (context === null)
            return;

        // Careful: this should be kept in sync with ContextBlockImg.
        // (so if this works on non-images that must work there too).
        if (context.targetURI === null) {
            this.label = '';
            return;
        }

        // Truncate to 100 characters. That is a silly "magic" number,
        // but this should only trigger on insanely long hostnames, so
        // making this a (hidden) pref seems like overkill.
        let host = context.targetURI.host;
        if (host.length > 100)
            host = _('image_server_truncate', host.slice(-100));

        this.label = _('image_server_target', host);
    }
};


/**
 * @class Context-sensitive submenu.
 *
 * @param {Object} childmap Mapping from layout name to Submenu instance.
 */
function ContextMenu(childmap) {
    // We initialize with a dummy layout. tweak() sets a sane one.
    LocalizedSubmenu.call(this, 'menu_context', []);
    this.map = childmap;
}

ContextMenu.prototype = {
    __proto__: LocalizedSubmenu.prototype,

    tweak: function(context) {
        if (context === null)
            return null;

        if (context.next === null)
            // No more contexts (currently unreached code because the
            // context chain ends in the scrollmenu which does not contain
            // a contextmenu child).
            return new menu.Dummy();
        let next = this.map[context.next.layout];
        if (next === undefined)
            // We have no layout for this context. This is probably an error,
            // but just hide it for now.
            return new menu.Dummy();
        // Pretend to be the next submenu. We cannot just return it because we
        // need to modify menu.context in activateMenu.
        this.label = next.label;
        this.iconname = next.iconname;
        this.layout = next.layout;
        return null;
    },

    activateMenu: function(piemenu) {
        piemenu.context.dehighlight();

        // This is a bit dirty: relies on nothing keeping a second
        // reference to context. That should be a valid assumption though.
        piemenu.context = piemenu.context.next;

        piemenu.context.highlight();

        // Forward to the menu we are pretending to be. This is
        // exactly right for the scrollmenu, which is currently the
        // only other user of activateMenu. May have to redesign this
        // if activateMenu gains more users than us and the
        // scrollmenu.
        this.map[piemenu.context.layout].activateMenu(piemenu);
    }
};


/**
 * @class Either the page or frame submenu depending on context.
 *
 * @param {Submenu} framemenu
 * @param {Submenu} pagemenu
 */
function ContextPage(framemenu, pagemenu) {
    // Submenu subclass to make the submenu arrow show up in the prefdialog.
    LocalizedSubmenu.call(this, 'context_page', []);
    this.framemenu = framemenu;
    this.pagemenu = pagemenu;
    this.active = false;
}

ContextPage.prototype = {
    __proto__: LocalizedSubmenu.prototype,

    tweak: function(context) {
        if (context === null)
            return null;
        else if (context.window.content.document !== context.frameDoc)
            return this.framemenu;
        else
            return this.pagemenu;
    }
};


/**
 * @class A stop button while loading, a reload button when done.
 *
 * @param {Action} reloadaction
 * @param {Action} stopaction
 */
function ContextStop(reloadaction, stopaction) {
    LocalizedAction.call(this, 'context_stop');
    this.reloadaction = reloadaction;
    this.stopaction = stopaction;
}

ContextStop.prototype = {
    __proto__: LocalizedAction.prototype,

    tweak: function(context) {
        // Make sure we show up with our "we are a context-sensitive action"
        // label when customizing:
        if (context === null)
            return null;

        this.reloadaction.tweak(context);
        if (this.reloadaction.active)
            return this.reloadaction;
        else
            return this.stopaction;
    }
};


/**
 * @class "block server" or "unblock server" depending on current state.
 *
 * Notice this will work in link instead of image context too.
 * Currently this is considered a feature.
 */
function ContextBlockImg() {
    LocalizedAction.call(this, 'context_blockimg');
    this.nsIPermissionManager = Components.interfaces.nsIPermissionManager;
    this.pmanager = Components.classes['@mozilla.org/permissionmanager;1']
        .getService(this.nsIPermissionManager);
}

ContextBlockImg.prototype = {
    __proto__: LocalizedAction.prototype,

    tweak: function(context) {
        // Careful: this should be kept in sync with ImageServerLabel.
        // (so if this works on non-images that must work there too).
        if (context === null || context.targetURI === null)
            return;

        // XXX refactor this to get rid of the _('foo_bar'), 'foo/bar' pair
        // (like we did with LocalizedAction/Submenu).
        if (this.pmanager.testPermission(context.targetURI, 'image') ===
            this.nsIPermissionManager.DENY_ACTION) {
            this.label = _('image_unblock');
            this.iconname = 'image/unblock';
        } else {
            this.label = _('image_block');
            this.iconname = 'image/block';
        }
    },

    doAction: function(context) {
        if (context.targetURI === null)
            return;

        let action;
        if (this.pmanager.testPermission(context.targetURI, 'image') ===
            this.nsIPermissionManager.DENY_ACTION)
            action = this.nsIPermissionManager.ALLOW_ACTION;
        else
            action = this.nsIPermissionManager.DENY_ACTION;
        this.pmanager.add(context.targetURI, 'image', action);
    }
};

/**
 * @class View source of the selection if there is one, everything if not.
 *
 * @param {String} name Label/icon used when customizing.
 * @param {Action} viewSourceAction
 * @param {Action} viewSelectionAction
 */
function ContextViewSource(label, viewSourceAction, viewSelectionAction) {
    LocalizedAction.call(this, label);
    this.viewSourceAction = viewSourceAction;
    this.viewSelectionAction = viewSelectionAction;
}

ContextViewSource.prototype = {
    __proto__: LocalizedAction.prototype,

    tweak: function(context) {
        if (context === null)
            return null;

        if (context.selection)
            return this.viewSelectionAction;
        else
            return this.viewSourceAction;
    }
};

/**
 * Base for the layouts reachable through the context chain (the ones
 * with a toplevel and submenu layout pair). Keep this and
 * _submenuNames in sync with the strings in labels.properties,
 * categories.properties and menus.properties.
 */
const _contextLayouts = {
    navigation: [
        'navigation_forward',
        'context_stop',
        'navigation_back',
        'navigation_up',
    ],
    link: [
        'link_in_newtab',
        'link_save',
        'link_in_newwin',
        'link_copy_url',
    ],
    mail: [
        'dummy',
        'dummy',
        'dummy',
        'link_copy_email',
    ],
    edit: [
        'edit_copy',
        'edit_paste',
        'edit_cut',
        'edit_undo',
    ],
    image: [
        'context_blockimg',
        'image_save',
        'image_view',
        'image_copy_url',
    ],
    selection: [
        'selection_open',
        'selection_translate',
        'selection_search',
        'selection_copy',
    ]
};

const _submenuNames = ['frame', 'page', 'window', 'tab']
    .concat([name for (name in _contextLayouts)]);
// Careful: JS "concat" takes either arrays or plain values, so this
// has to be a listcomp, not a genexp (the genexp itself gets added)

/**
 * Other half of the menu for toplevel _contextLayouts-based menus:
 */
const _toplevelLayoutBase = [
    'menu_tab',
    'context_page',
    'context_2',
    'menu_window',
];

const _submenuLayoutBase = ['context_2', 'dummy', 'dummy', 'dummy'];

/**
 * @class Container for the actions and layouts mappings.
 * May also act as a preference observer.
 * Remember some of the actions in this.actions are stateful,
 * so do not use the same ActionManager for more than one menu
 * simultaneously.
 * @implements nsIObserver
 *
 * @param {nsIPrefBranch2} prefbranch Branch for the "customizations" tree.
 *   May be null.
 */
function ActionManager(prefbranch) {
    // We could easily create a prefbranch here instead of getting one
    // passed in, but then we would have to store a reference to it so
    // we can unregister the listener. Since the main user
    // (integration.js) already keeps a reference to the right
    // prefbranch anyway we do it this way.

    let actions = {};
    for (let [name, func] in Iterator(_simpleActions)) {
        let action = new LocalizedAction(name);
        action.doAction = func;
        actions[name] = action;
    }
    for (let [actionname, commandname] in Iterator(_commandActions))
        actions[actionname] = new CommandAction(actionname, commandname);

    for (let [actionname, commandname] in Iterator(_controllerCommandActions))
        actions[actionname] = new ControllerCommandAction(actionname,
                                                          commandname);

    for (let [name, func] in Iterator(_simpleContextMenuActions)) {
        let action = actions[name] = new ContextMenuAction(name);
        action.doMenuAction = func;
    }

    // Dummies.
    actions.dummy = new menu.Dummy();
    // options.js needs name set:
    actions.dummy.name = 'dummy';
    actions.image_server = new ImageServerLabel();

    // Actions.
    actions.navigation_forward = new NaviForward();
    actions.navigation_back = new NaviBack();

    actions.context_stop = new ContextStop(actions.navigation_reload,
                                           actions.navigation_stop);
    actions.context_blockimg = new ContextBlockImg();
    actions.context_page_source = new ContextViewSource(
        'context_page_source', actions.page_source, actions.selection_source);
    actions.context_frame_source = new ContextViewSource(
        'context_frame_source',
        actions.frame_source, actions.selection_source);

    actions.selection_search = new SearchAction(false);
    actions.selection_search_ntab = new SearchAction(true);

    // Submenus.

    // HACK: instead of directly constructing the Submenu actions (which
    // is nontrivial because of nested references, sometimes involving
    // context-sensitive actions too) we pass an empty layout here, and
    // swap it for the actual layout later. This works because nothing
    // except for the actual PieMenu accesses Submenu.layout.

    for each (let name in _submenuNames)
        actions['menu_' + name] = new LocalizedSubmenu('menu_' + name, []);

    // HACK: monkeypatch activateMenu for the frame. This code is
    // probably temporary, and I do not yet know how the final version
    // will be factored.
    actions.menu_frame.activateMenu = function(piemenu) {
        // HACK: we want to highlight the frame we operate on, but currently
        // highlighting is tied to the context and we do not have a separate
        // context. Currently we just hack up that context:
        piemenu.context.dehighlight();
        piemenu.context.target = piemenu.context.frameDoc.body;
        piemenu.context.highlight();
    };

    actions.context_page = new ContextPage(actions.menu_frame,
                                           actions.menu_page);
    actions.menu_scroll = new ScrollMenu();

    let contextMenuMap = {'scroll': actions.menu_scroll};
    for (let name in _contextLayouts)
        contextMenuMap[name] = actions['menu_' + name];
    actions.context_2 = new ContextMenu(contextMenuMap);

    // Alternate actions.
    actions.navigation_reload.altaction = actions.navigation_reload_skip;
    actions.navigation_forward.altaction = actions.navigation_lastpage;
    actions.navigation_back.altaction = actions.navigation_firstpage;
    actions.navigation_up.altaction = actions.navigation_enter_address;
    actions.tab_new.altaction = actions.tab_duplicate;
    actions.tab_close.altaction = actions.tab_close_others;
    actions.window_new.altaction = actions.window_duplicate;
    actions.window_close.altaction = actions.window_close_others;
    actions.window_minimize.altaction = actions.window_togglebar;
    actions.page_info.altaction = actions.page_validate;
    actions.selection_open.altaction = actions.selection_open_ntab;
    actions.selection_search.altaction = actions.selection_search_ntab;
    actions.selection_translate.altaction = actions.selection_translate_ntab;

    this.actions = actions;

    if (prefbranch !== null)
        prefbranch.addObserver('customizations.', this, false);

    this._updateLayouts(prefbranch);
}

ActionManager.prototype = {
    /**
     * Unregister pref observer.
     *
     * @param {nsIPrefBranch2} prefbranch
     *   The branch passed to the constructor.
     */
    shutdown: function(prefbranch) {
        prefbranch.removeObserver('customizations.', this);
    },

    /**
     * Parse a pref key.
     *
     * @param {nsIPrefBranch2} prefbranch Branch with "customizations" key.
     * @param {String} key Key like 'customizations.blah.2'.
     * @param {Bool} allowNullValue Whether to return null if key is unset.
     * @returns [layoutName, index, value] or null
     */
    parseKey: function(prefbranch, key, allowNullValue) {
        let split = key.split('.');
        if (split.length !== 3)
            return null;
        // The first value is "customizations" which we ignore.
        let [, layoutName, index] = split;

        if (!this.layouts.hasOwnProperty(layoutName))
            return null;

        // Sanity check preventing multiple prefs specifing the same index:
        if (!/^(0|[1-9][0-9]*)$/.test(index))
            return null;
        index = parseInt(index);
        if (index >= this.layouts[layoutName].length)
            return null;

        // Ok, the key is valid.
        try {
            var value = prefbranch.getCharPref(key);
        } catch (e) {
            // Assume this failed because the pref is not set.
            if (allowNullValue)
                return [layoutName, index, null];
            else
                return null;
        }
        if (!this.actions.hasOwnProperty(value))
            return null;

        // Value is valid too, store:
        return [layoutName, index, value];
    },

    /**
     * Update this.layouts.
     */
    _updateLayouts: function(prefbranch) {
        // The way this works with updates is hacky but effective. First,
        // updating the array in place works because nothing keeps a
        // longterm reference to a "layout" array: the menu copies it
        // (applying tweakfunctions) on initialization and when displaying
        // a submenu. And we have to do a full reinitialization on pref
        // change because we do get notified about the change but do not
        // get the previous value, so unless we manually track the applied
        // customizations we cannot apply just the changes.
        let actions = this.actions;

        this.layouts = {
            // Submenus, direct children of the main menu:
            tab_sub: [
                actions.dummy,            actions.tab_next,
                actions.dummy,            actions.tab_close,
                actions.tab_undo_close,   actions.tab_prev,
                actions.dummy,            actions.tab_new,
            ],
            window_sub: [
                actions.dummy,            actions.window_fullscreen,
                actions.dummy,            actions.window_close,
                actions.dummy,            actions.window_minimize,
                actions.dummy,            actions.window_new,
            ],
            // Either page or frame is also a child of the main menu:
            page_sub: [
                actions.dummy,            actions.page_bookmark,
                actions.dummy,            actions.page_save,
                actions.dummy,            actions.context_page_source,
                actions.dummy,            actions.page_info,
            ],
            frame_sub: [
                actions.dummy,            actions.frame_view,
                actions.frame_reload,     actions.frame_save,
                actions.dummy,            actions.context_frame_source,
                actions.menu_page,        actions.frame_info,
            ]
        };

        for (let [name, halfLayout] in Iterator(_contextLayouts)) {
            // Construct the toplevel menu layout:
            let layout = [];
            for (let i = 0; i < 4; ++i) {
                layout.push(this.actions[_toplevelLayoutBase[i]]);
                layout.push(this.actions[halfLayout[i]]);
            }
            this.layouts[name] = layout;

            // Construct the submenu layout:
            layout = [];
            for (let i = 0; i < 4; ++i) {
                layout.push(this.actions[_submenuLayoutBase[i]]);
                layout.push(this.actions[halfLayout[i]]);
            }
            this.layouts[name + '_sub'] = layout;
        }

        // The image submenu needs a weird special case for its label:
        this.layouts.image_sub[2] = actions.image_server;

        // Apply customizations:
        if (prefbranch !== null) {
            for each (let key in prefbranch.getChildList('customizations.',
                                                         {})) {
                let entry = this.parseKey(prefbranch, key, false);
                if (entry !== null) {
                    let [layoutName, index, value] = entry;
                    this.layouts[layoutName][index] = actions[value];
                }
            }
        }

        // Now fix up the submenus as described above:
        for each (let submenu in _submenuNames)
            actions['menu_' + submenu].layout =
                this.layouts[submenu + '_sub'];
    },

    /**
     * Callback for the pref service.
     */
    observe: function(prefbranch, topic, data) {
        if (topic !== 'nsPref:changed')
            throw new Error('huh, I did not sign up for ' + topic);

        // This is required, without it the prefbranch only shows up as
        // nsISupports.
        prefbranch.QueryInterface(Components.interfaces.nsIPrefBranch);
        this._updateLayouts(prefbranch);
    }
};
